/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011-2012 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stdlib.h>
#include <string.h>
#include "vi3x_io.h"
#include "vi3x_worker.h"
#include "vi3x_modules.h"

static char const *Tag="vi3x";

enum OPT_ID
    {
    OPT_BAD=-1,
    OPT_HELP,
    OPT_FULL_HELP,
    OPT_IN,
    OPT_OUT,
    OPT_FILTER,
    OPT_LOG,
    OPT_TEST,
    OPT_COUNT
    };

typedef struct opt_map_s
    {
    char *opt;
    enum OPT_ID id;
    } opt_map_t;

static opt_map_t const OptsHelp[]=
    {
    {"-h", OPT_HELP},
    {"--help", OPT_HELP},
    {NULL, OPT_BAD}
    };

static opt_map_t const OptsFullHelp[]=
    {
    {"--full-help", OPT_FULL_HELP},
    {"--fullhelp", OPT_FULL_HELP},
    {"--full_help", OPT_FULL_HELP},
    {NULL, OPT_BAD}
    };

static opt_map_t const OptsLog[]=
    {
    {"-l", OPT_LOG},
    {"--log", OPT_LOG},
    {NULL, OPT_BAD}
    };

static opt_map_t const Opts[]=
    {
    {"-i", OPT_IN},
    {"--input", OPT_IN},
    {"-o", OPT_OUT},
    {"--output", OPT_OUT},
    {"-f", OPT_FILTER},
    {"--filter", OPT_FILTER},
    {"-t", OPT_TEST},
    {"--test", OPT_TEST},
    {NULL, OPT_BAD}
    };

static char const *UsageText[]=
    {
    "video3x version " VI3X_VER_STR " LGPL 3+\n",
    "Copyright 2011-2012 Vadim Druzhin <cdslow@mail.ru>\n"
    "\n"
    "Call: vi3x -i in[:arg] -o out[:arg] [-f filter[@frames][:arg] ...] [options]\n"
    "  or: vi3x -h [-i in ...] [-o out ...] [-f filter ...] [options]\n"
    "  or: vi3x --full-help [options]\n"
    "\n",
    "SYNTAX:\n"
    "in - specifies an input module (i.e., 420p).\n"
    "out - specifies an output module (i.e., null).\n"
    "filter - specifies a filter module (i.e., timecode).\n"
    "arg - presents a list of module parameters in the form:\n"
    "    name1=value1,name2=value2 and so on.\n"
    "    If a value is 1, it may be omitted (name2 means name2=1).\n"
    "frames - presents a list of frame ranges in the form: f1-f2,f3-f4 and so on.\n",
    "    The frame numbering starts from 1, ranges are inclusive\n"
    "    (11-20 means process 10 frames including frames 11 and 20).\n"
    "    The 1-frame range may be specified with a single number\n"
    "    (11-11 is equal to just 11).\n"
    "    The frames are counted for each filter individually, so if some filter\n"
    "    in the chain changes a frame rate (adds or drops frames)\n"
    "    the frame numbering will be shifted. I.e., after an inverse telecine filter\n"
    "    the 30th frame becomes the 24th frame for consecutive filters.\n"
    "\n",
    "Full vi3x command line may look like this:\n"
    "vi3x -i 420p:w=640,h=480 -o 420p -f toonclean:demo -f timecode@100-200:x=10,y=10\n"
    "\n",
    NULL
    };

static opt_map_t const OptsDesc[]=
    {
    {
    " - display this help text.\n"
    "    If you need a help for specific module, add the module specification\n"
    "    to the command line: vi3x -h -o 420p -o null -f timecode\n",
    OPT_HELP
    },
    {
    " - display the help for all modules.\n",
    OPT_FULL_HELP
    },
    {
    " in[:arg] - select the input module. This is a mandatory option.\n",
    OPT_IN
    },
    {
    " out[:arg] - select the output module. This is a mandatory option.\n",
    OPT_OUT
    },
    {
    " filter[@frames][:arg] - append the filter module\n"
    "    to a filter chain. This option may be used multiple times.\n",
    OPT_FILTER
    },
    {
    " log_file - log all messages to a file.\n",
    OPT_LOG
    },
    {
    " - initialize the filter chain, but don't process any input.\n",
    OPT_TEST
    },
    {NULL, OPT_BAD}
    };

static enum OPT_ID opt_match(opt_map_t const *map, char const *opt);
static char **argv_match(opt_map_t const *map, char **argv);
static void opt_help(opt_map_t const *desc, opt_map_t const *map);
static void usage(void);
static void full_help(void);
static void usage_general(void);
static int do_vi3x(char **opt, vi3x_workflow_t *wf);

int main(int argc, char *argv[])
    {
    vi3x_workflow_t *wf=NULL;
    char **opt;
    int ret=EXIT_SUCCESS;

    vi3x_register_modules();

    if(argc < 2)
        {
        vi3x_err("[%s] Error: Missing mandatory options!\n", Tag);
        usage();
        return EXIT_FAILURE;
        }

    /* Redirect log */
    opt=argv_match(OptsLog, argv + 1);
    if(opt)
        {
        if(NULL == opt[1])
            {
            vi3x_err("[%s] Error: Missing log file name!\n", Tag);
            return EXIT_FAILURE;
            }
        if(!vi3x_set_log(opt[1]))
            return EXIT_FAILURE;
        else
        *opt[0]=0;
        *opt[1]=0;
        }

    if(argv_match(OptsFullHelp, argv+1))
        full_help(); /* Do full help */
    else
        {
        /* Check help flag */
        opt=argv_match(OptsHelp, argv + 1);
        if(opt)
            {
            usage();
            *opt[0]=0;
            }
        else
            wf=vi3x_create();

        ret=do_vi3x(argv + 1, wf);
        }


    if(wf)
        vi3x_destroy(wf);

    vi3x_set_log(NULL);

    return ret;
    }

static int do_vi3x(char **opt, vi3x_workflow_t *wf)
    {
    int test=0;

    while(NULL != *opt)
        {
        switch(opt_match(Opts, *opt))
            {
        case OPT_IN:
            if(NULL==opt[1])
                {
                vi3x_err("[%s] Error: Missing input module name!\n", Tag);
                return EXIT_FAILURE;
                }
            if(NULL == wf)
                vi3x_help_import(opt[1]);
            else if(!vi3x_add_import(wf, opt[1]))
                return EXIT_FAILURE;
            ++opt;
            break;
        case OPT_OUT:
            if(NULL == opt[1])
                {
                vi3x_err("[%s] Error: Missing output module name!\n", Tag);
                return EXIT_FAILURE;
                }
            if(NULL == wf)
                vi3x_help_export(opt[1]);
            else if(!vi3x_add_export(wf, opt[1]))
                return EXIT_FAILURE;
            ++opt;
            break;
        case OPT_FILTER:
            if(NULL == opt[1])
                {
                vi3x_err("[%s] Error: Missing filter module name!\n", Tag);
                return EXIT_FAILURE;
                }
            if(NULL == wf)
                vi3x_help_filter(opt[1]);
            else if(!vi3x_add_filter(wf, opt[1]))
                return EXIT_FAILURE;
            ++opt;
            break;
        case OPT_TEST:
            test=1;
            break;
        default:
            if(0 != **opt)
                {
                vi3x_err("[%s] Error: Invalid option: '%s'!\n", Tag, *opt);
                return EXIT_FAILURE;
                }
            }
        ++opt;
        }

    if(wf)
        {
        if(!vi3x_init(wf))
            return EXIT_FAILURE;

        if(!test)
            {
            if(!vi3x_run(wf))
                return EXIT_FAILURE;
            }
        }

    return EXIT_SUCCESS;
    }

static void usage(void)
    {
    usage_general();
    vi3x_help("INPUT MODULES:\n");
    vi3x_help_all_import(1);
    vi3x_help("\nOUTPUT MODULES:\n");
    vi3x_help_all_export(1);
    vi3x_help("\nFILTER MODULES:\n");
    vi3x_help_all_filter(1);
    vi3x_help("\n");
    }

static void full_help(void)
    {
    usage_general();
    vi3x_help("*** INPUT MODULES:\n\n");
    vi3x_help_all_import(0);
    vi3x_help("*** OUTPUT MODULES:\n\n");
    vi3x_help_all_export(0);
    vi3x_help("*** FILTER MODULES:\n\n");
    vi3x_help_all_filter(0);
    }

static void usage_general(void)
    {
    char const **txt;

    for(txt=UsageText; NULL != *txt; ++txt)
        vi3x_help(*txt);
    vi3x_help("OPTIONS:\n");
    opt_help(OptsDesc, OptsHelp);
    opt_help(OptsDesc, OptsFullHelp);
    opt_help(OptsDesc, OptsLog);
    opt_help(OptsDesc, Opts);
    vi3x_help("\n");
    }

static enum OPT_ID opt_match(opt_map_t const *map, char const *opt)
    {
    while(NULL != map->opt)
        {
        if(0 == strcmp(map->opt, opt))
            return map->id;
        ++map;
        }

    return OPT_BAD;
    }

static char **argv_match(opt_map_t const *map, char **argv)
    {
    while(NULL != *argv)
        {
        if(OPT_BAD != opt_match(map, *argv))
            return argv;
        ++argv;
        }

    return NULL;
    }

static void opt_help(opt_map_t const *desc, opt_map_t const *map)
    {
    int count[OPT_COUNT];
    opt_map_t const *m;
    int i;
    
    memset(count, 0, sizeof(count));
    for(m=map; m->id >= 0 && m->id < OPT_COUNT; ++m)
        ++count[m->id];

    for(i=0; i < OPT_COUNT; ++i)
        {
        if(0==count[i])
            continue;
        for(m=map; NULL != m->opt; ++m)
            {
            if(m->id == i)
                {
                vi3x_help("%s", m->opt);
                --count[i];
                if(count[i] > 0)
                    vi3x_help("|");
                }
            }
        for(m=desc; NULL != m->opt; ++m)
            {
            if(m->id == i)
                vi3x_help("%s", m->opt);
            }
        }
    }
