/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011-2012 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stddef.h>
#include <string.h>
#include "vi3x_modules.h"

static char const *Tag="420p";

/* Instance local data */
typedef struct m_priv_s
    {
    int mpeg1;
    int il;
    int no_il;
    vi3x_420_t i420;
    } m_priv_t;

static int init_420p(vi3x_module_t *m);
static int fini_420p(vi3x_module_t *m);
static int image_420p(vi3x_module_t *m, vi3x_image_t *image);
static void help_420p(vi3x_help_t *h);
static void help_mpeg1(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_il(vi3x_arg_t const *arg, vi3x_help_t const *h);

/* Module arguments */
static vi3x_arg_t ArgMap[]=
    {
    { "mpeg1", 0, offsetof(m_priv_t, mpeg1), help_mpeg1 },
    { "il",   -1, offsetof(m_priv_t, il), help_il },
    { NULL, 0, 0, NULL }
    };

/* Entry point */
VI3X_MAIN()
    {
    vi3x_reg_export(Tag, init_420p, sizeof(m_priv_t), help_420p);
    }

/* Module help */
static void help_420p(vi3x_help_t *h)
    {
    h->help_module(Tag, VI3X_VER_STR, "Vadim Druzhin",
        "Export raw frames in the planar 420p color space",
        "The frames are written to the standard output.",
        ArgMap);
    }

static void help_mpeg1(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Sample position of downsampled chroma planes", "",
        arg->def,
        "- assume MPEG-2 sample position;",
        "- assume MPEG-1 sample position."
        );
    }

static void help_il(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_list(arg->name,
        "Interlacing of downsampled chroma planes", "",
        arg->def, 3,
        -1, "- use the frame interlacing flag;",
        0, " - force progressive downsampling;",
        1, " - force interlaced downsampling."
        );
    }

/* Instance initialization */
static int init_420p(vi3x_module_t *m)
    {
    /* Parse and verify arguments */
    if(!m->vi3x.parse_args(m->priv, m->args, ArgMap, Tag))
        return VI3X_RET_ERR;

    if(m->w <= 0 || m->h <= 0)
        {
        m->vi3x.err("[%s] Error: Incorrect frame size: %d x %d !\n",
            Tag, m->w, m->h);
        return VI3X_RET_ERR;
        }

    if((m->w & 1) || (m->h & 1))
        {
        m->vi3x.err("[%s] Error: Uneven frame size: %d x %d !\n",
            Tag, m->w, m->h);
        return VI3X_RET_ERR;
        }

    m->priv->no_il=m->h & 3;
    if(m->priv->no_il)
        {
        if(1 == m->priv->il)
            {
            m->vi3x.err("[%s] Error: Height must be multiple of 4 "
                "for interlacing: h=%d !\n", Tag, m->h);
            return VI3X_RET_ERR;
            }
        else if(0!=m->priv->il)
            m->vi3x.warn("[%s] Warning: Interlacing impossible (h=%d).\n",
                Tag, m->h);
        }

    /* Initialize output image buffer */
    m->vi3x.init_420(&m->priv->i420, m->w, m->h, NULL);
    m->vi3x.init_420(&m->priv->i420, m->w, m->h,
        m->vi3x.alloc(m->priv->i420.img_sz, Tag));

    /* Publish module interface */
    m->fini = fini_420p;
    m->process_image = image_420p;

    return VI3X_RET_OK;
    }

/* Instance finalization */
static int fini_420p(vi3x_module_t *m)
    {
    /* Free instance memory */
    m->vi3x.free(m->priv->i420.yy);

    return VI3X_RET_OK;
    }

/* Frame processor */
static int image_420p(vi3x_module_t *m, vi3x_image_t *image)
    {
    int ret;

    /* Overwrite image type */
    if(-1 != m->priv->il)
        image->il=m->priv->il;

    if(image->il && m->priv->no_il)
        {
        m->vi3x.err("[%s] Error: Can't process interlaced frame!\n", Tag);
        return 0;
        }

    /* Downsample frame for output */
    if(m->priv->mpeg1)
        m->vi3x.downsample_mpeg1(&(m->priv->i420), image);
    else
        m->vi3x.downsample_mpeg2(&(m->priv->i420), image);

    /* Write whole frame, color planes order assumed as in default vi3x_420_t */
    ret=m->vi3x.write(m->priv->i420.yy, m->priv->i420.img_sz);
    if(ret != m->priv->i420.img_sz)
        {
        m->vi3x.err("[%s] Error: Frame %ld - write failed!\n", Tag, m->frame_n);
        return 0;
        }

    return 1;
    }
