/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2012 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stdint.h>
#include <string.h>
#include <stddef.h>
#include <limits.h>
#include <stdlib.h>
#include "vi3x_modules.h"

static char const *Tag = "chromalock";

#define Q_DEPTH 3
typedef struct m_priv_s
    {
    int motion;
    int show;
    vi3x_image_t frames[Q_DEPTH];
    int q_len;
    int q_in;
    int q_out;
    int q_fill;
    } m_priv_t;

#define MARK_YY VI3X_MAX
#define MARK_CB VI3X_GREY
#define MARK_CR VI3X_GREY

static int m_init(vi3x_module_t *m);
static int m_fini(vi3x_module_t *m);
static int m_process(vi3x_module_t *m, vi3x_image_t *image);
static int m_flush(vi3x_module_t *m, vi3x_image_t *image);
static void reset_q(m_priv_t *p);
static int do_frame(vi3x_module_t *m, vi3x_image_t *image);
static void m_help(vi3x_help_t *h);
static void blend_chroma(vi3x_image_t *img,
    vi3x_image_t const frames[Q_DEPTH], int n, int motion, int show);
static inline int sqd3(int px1, int px, int px2);
static inline void blend_xy(
    vi3x_image_t *img, vi3x_image_t const frames[Q_DEPTH], int n, int xy);
static inline int is_noise(int px1, int px, int px2);
static inline int avg3(int px1, int px, int px2);
static void help_motion(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_show(vi3x_arg_t const *arg, vi3x_help_t const *h);

static const vi3x_arg_t ArgMap[] =
    {
    { "motion", 32, offsetof(m_priv_t, motion), help_motion },
    { "show",    0, offsetof(m_priv_t, show), help_show },
    { NULL, 0, 0, NULL }
    };

VI3X_MAIN()
    {
    vi3x_reg_filter(Tag, m_init, sizeof(m_priv_t), m_help);
    }

static int m_init(vi3x_module_t *m)
    {
    vi3x_sample_t *buf;
    int i;

    /* Initialize module parameters */
    if(!m->vi3x.parse_args(m->priv, m->args, ArgMap, Tag))
        return VI3X_RET_ERR;

    /* Calculate image buffer size */
    m->vi3x.init_image(&m->priv->frames[0], m->w, m->h, NULL);

    /* Allocate buffer */
    buf = m->vi3x.alloc(m->priv->frames[0].img_sz * Q_DEPTH, Tag);

    /* Initialize buffer pointers */
    for(i = 0; i < Q_DEPTH; ++i)
        buf = m->vi3x.init_image(m->priv->frames + i, m->w, m->h, buf);

    /* Initialize state */
    reset_q(m->priv);

    /* Publish module interface */
    m->fini = m_fini;
    m->process_image = m_process;
    m->flush = m_flush;

    return VI3X_RET_OK;
    }

/* Prepare queue */
static void reset_q(m_priv_t *p)
    {
    p->q_in = 0;
    p->q_out = 0;
    p->q_len = 0;
    p->q_fill = 0;
    }

static int m_fini(vi3x_module_t *m)
    {
    m->vi3x.free(m->priv->frames[0].yy);

    return VI3X_RET_OK;
    }

static int m_process(vi3x_module_t *m, vi3x_image_t *image)
    {
    m_priv_t *p = m->priv;

    /* Put next frame into queue */
    m->vi3x.copy_image(p->frames + p->q_in, image);
    p->q_in=(p->q_in + 1) % Q_DEPTH;
    ++p->q_len;
    ++p->q_fill;

    /* If queue ready */
    if(Q_DEPTH == p->q_fill)
        return do_frame(m, image); /* Process queued frames */

    if(1 == p->q_len)
        {
        /* Duplicate first frame */
        m->vi3x.copy_image(p->frames + p->q_in, image);
        ++p->q_in;
        ++p->q_out;
        ++p->q_fill;
        }

    return VI3X_RET_OK;
    }

static int m_flush(vi3x_module_t *m, vi3x_image_t *image)
    {
    m_priv_t *p = m->priv;
    int ret = VI3X_RET_OK;

    /* While there is frames in the queue */
    while(ret && p->q_len > 0)
        {
        /* Duplicate last frame to fill up the queue */
        while(p->q_fill < Q_DEPTH)
            {
            m->vi3x.copy_image(p->frames + p->q_in,
                p->frames + (p->q_in - 1 + Q_DEPTH) % Q_DEPTH);
            p->q_in = (p->q_in + 1) % Q_DEPTH;
            ++p->q_fill;
            }
        /* Process queued frames */
        ret = do_frame(m, image);
        }

    /* Empty the queue */
    reset_q(m->priv);

    return ret;
    }

/* Filter queued frames */
static int do_frame(vi3x_module_t *m, vi3x_image_t *image)
    {
    m_priv_t *p = m->priv;

    /* Process images */
    blend_chroma(image, p->frames, p->q_out, p->motion, p->show);

    /* Assign correct il flag to output image */
    image->il = p->frames[p->q_out].il;

    /* Remove first frame from queue */
    p->q_out = (p->q_out + 1) % Q_DEPTH;
    --p->q_len;
    --p->q_fill;

    /* Output processed frame */
    return m->out_image(m, image);
    }

/* Blend chroma values in the areas with static luma */
static void blend_chroma(vi3x_image_t *img,
    vi3x_image_t const frames[Q_DEPTH], int n, int motion, int show)
    {
    int xy;

    /* Pixel by pixel */
    for(xy = 0; xy < img->count; ++xy)
        {
        /* Estimate the motion */
        int e = sqd3(frames[0].yy[xy], frames[1].yy[xy], frames[2].yy[xy]);
        e = e / 2; /* Normalize to 0-255**2 */

        if(!show && e < motion)
            /* No motion, process the chroma */
            blend_xy(img, frames, n, xy);
        else if(show && e >= motion)
            {
            /* Mark areas with high motion */
            img->yy[xy] = MARK_YY;
            img->cb[xy] = MARK_CB;
            img->cr[xy] = MARK_CR;
            }
        else
            {
            /* Copy the pixel as is */
            img->yy[xy] = frames[n].yy[xy];
            img->cb[xy] = frames[n].cb[xy];
            img->cr[xy] = frames[n].cr[xy];
            }
        }
    }

/* Sum of quadratic difference */
static inline int sqd3(int px1, int px, int px2)
    {
    px1 -= px;
    px2 -= px;

    return px1 * px1 + px2 * px2;
    }

/* Blend fast changing (noisy) pixels */
static inline void blend_xy(
    vi3x_image_t *img, vi3x_image_t const frames[Q_DEPTH], int n, int xy)
    {
    int n1 = (n + 1) % Q_DEPTH; /* Next frame number */
    int n2 = (n + Q_DEPTH - 1) % Q_DEPTH; /* Previous frame number */

    /* Blue plane */
    if(is_noise(frames[n1].cb[xy], frames[n].cb[xy], frames[n2].cb[xy]))
        /* Noisy pixel, average it */
        img->cb[xy] = avg3(
            frames[n1].cb[xy], frames[n].cb[xy], frames[n2].cb[xy]);
    else
        /* Stable pixel, copy as is */
        img->cb[xy] = frames[n].cb[xy];

    /* Red plane */
    if(is_noise(frames[n1].cr[xy], frames[n].cr[xy], frames[n2].cr[xy]))
        /* Noisy pixel, average it */
        img->cr[xy] = avg3(
            frames[n1].cr[xy], frames[n].cr[xy], frames[n2].cr[xy]);
    else
        /* Stable pixel, copy as is */
        img->cr[xy] = frames[n].cr[xy];

    /* Copy the luma as is */
    img->yy[xy] = frames[n].yy[xy];
    }

static inline int is_noise(int px1, int px, int px2)
    {
    int ad12 = abs(px1 - px2) * 2;

    return ad12 < abs(px - px1) && ad12 < abs(px - px2);
    }

static inline int avg3(int px1, int px, int px2)
    {
    return (px * 2 + px1 + px2 + 2) / 4;
    }

/* Module help */
static void m_help(vi3x_help_t *h)
    {
    h->help_module(Tag, VI3X_VER_STR, "Vadim Druzhin",
        "Temporal chroma stabilizer",
        "This filter performs temporal blending of chroma planes\n"
        "in static image areas. It is suitable for removing a 'rainbow'\n"
        "(extraneous moving colors in areas of fine luminance detail\n"
        "caused by composite video processing).",
        ArgMap
        );
    }

static void help_motion(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Motion detection sensitivity",
        "With smaller values more areas will be considered as moving\n"
        "and left intact by the filter.",
        arg->def, 0, VI3X_MAX * VI3X_MAX
        );
    }

static void help_show(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Show moving areas instead of performing a chroma blending", "",
        arg->def,
        "- Perform a chroma blending",
        "- Mark moving areas"
        );
    }
