/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2012 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stdint.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include "vi3x_modules.h"

static char const *Tag = "edgeclean";

/* Default edge levels */
#define EDGE_L ((VI3X_MAX + 1) * 3 / 8)
#define EDGE_C ((VI3X_MAX + 1) / 16)

/* Edge search interval */
#define EDGE_DX 2
#define EDGE_STEP (EDGE_DX * 2 + 1)

/* Denoise radius */
#define SKIP_DX 1
#define SKIP_DY 1

#define CONST_DX (EDGE_DX > SKIP_DX ? EDGE_DX : SKIP_DX)

/* Mark color */
#define SHOW_YY VI3X_MAX

typedef struct m_priv_s
    {
    int denoise;
    int edge_l;
    int edge_c;
    int noise_width;
    int line_width;
    int const_l;
    int show;
    vi3x_sample_t *tmp;
    } m_priv_t;

static void m_help(vi3x_help_t *h);
static void help_denoise(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_edge_l(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_edge_c(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_noise_width(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_line_width(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_const_l(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_show(vi3x_arg_t const *arg, vi3x_help_t const *h);

static const vi3x_arg_t ArgMap[] =
    {
    { "denoise",     2, offsetof(m_priv_t, denoise), help_denoise },
    { "edge_l", EDGE_L, offsetof(m_priv_t, edge_l), help_edge_l },
    { "edge_c", EDGE_C, offsetof(m_priv_t, edge_c), help_edge_c },
    { "noise_width", 4, offsetof(m_priv_t, noise_width), help_noise_width },
    { "line_width", 16, offsetof(m_priv_t, line_width), help_line_width },
    { "const_l",       0, offsetof(m_priv_t, const_l), help_const_l },
    { "show",        0, offsetof(m_priv_t, show), help_show },
    { NULL, 0, 0, NULL }
    };

static int m_init(vi3x_module_t *flt);
static int m_fini(vi3x_module_t *flt);
static int m_process(vi3x_module_t *flt, vi3x_image_t *image);

static void clean_edges(vi3x_image_t *img, m_priv_t *p, int edge_l, int show);
static inline int abs_diff(int a, int b);
static int clean_part(vi3x_sample_t *dst, vi3x_image_t const *img,
    int l, int r, int edge_l, int edge_c, int noise_width);
static inline void cond_avg(vi3x_sample_t *dst,
    vi3x_image_t const *img, int xy, int edge_l, int edge_c);
static int search_edge(
    vi3x_image_t const *img, int from, int to, int edge_l, int edge_c);
static int search_right(vi3x_image_t const *img,
    int from, int to, int edge_l, int edge_c, int line_width);
static int search_single_left(
    vi3x_sample_t const *px, int from, int to, int edge_level);
static int search_single_right(vi3x_sample_t const *px,
    int from, int to, int edge_level, int line_width);
static inline void edge_avg(vi3x_sample_t *dst,
    vi3x_image_t const *img, int xy, int edge_l, int edge_c);
static inline int avg_3(vi3x_sample_t const *px, int w);
static inline int avg_5(vi3x_sample_t const *px, int w);
static inline int pix_match(
    vi3x_image_t const *img, int xy1, int xy2, int el, int ec);

VI3X_MAIN()
    {
    vi3x_reg_filter(Tag, m_init, sizeof(m_priv_t), m_help);
    }

static int m_init(vi3x_module_t *m)
    {
    /* Initialize parameters */
    if(!m->vi3x.parse_args(m->priv, m->args, ArgMap, Tag))
        return VI3X_RET_ERR;

    if(m->priv->denoise < 1)
        {
        m->vi3x.err("[%s] Error: 'denoise' too small (%d < 1)!\n",
            Tag, m->priv->denoise);
        return VI3X_RET_ERR;
        }

    if(m->priv->edge_l < VI3X_MIN || m->priv->edge_l > VI3X_MAX)
        {
        m->vi3x.err("[%s] Error: 'edge_l' out of range (%d not in %d ... %d)!\n",
            Tag, m->priv->edge_l, VI3X_MIN, VI3X_MAX);
        return VI3X_RET_ERR;
        }

    if(m->priv->edge_c < VI3X_MIN || m->priv->edge_c > VI3X_MAX)
        {
        m->vi3x.err("[%s] Error: 'edge_c' out of range (%d not in %d ... %d)!\n",
            Tag, m->priv->edge_c, VI3X_MIN, VI3X_MAX);
        return VI3X_RET_ERR;
        }

    if(m->priv->noise_width < 0)
        {
        m->vi3x.err("[%s] Error: 'noise_width' can't be negative!\n", Tag);
        return VI3X_RET_ERR;
        }

    if(m->priv->line_width < 0)
        {
        m->vi3x.err("[%s] Error: 'line_width' can't be negative!\n", Tag);
        return VI3X_RET_ERR;
        }

    /* Allocate memory */
    m->priv->tmp = m->vi3x.alloc(m->w * m->h * sizeof(*m->priv->tmp), Tag);

    /* Initialize methods */
    m->process_image = m_process;
    m->fini = m_fini;

    return VI3X_RET_OK;
    }

static int m_fini(vi3x_module_t *m)
    {
    m->vi3x.free(m->priv->tmp);

    return VI3X_RET_OK;
    }

static int m_process(vi3x_module_t *m, vi3x_image_t *img)
    {
    int i;
    int edge_l = m->priv->edge_l;

    /* Copy orginal luma plane */
    memcpy(m->priv->tmp, img->yy, img->plane_sz);

    /* Clean edges multiple times */
    for(i = 0; i < m->priv->denoise - 1; ++i)
        {
        clean_edges(img, m->priv, edge_l, 0);
        /* Lower luma edge level for next iteration */
        if(!m->priv->const_l && edge_l >= m->priv->edge_c * 4)
            edge_l /= 2;
        }

    /* Show edges on last iteration (if desired) */
    clean_edges(img, m->priv, edge_l, m->priv->show);

    return m->out_image(m, img);
    }

/* Do the trick, original luma plane must be copied to p->tmp beforehand */
static void clean_edges(vi3x_image_t *img, m_priv_t *p, int edge_l, int show)
    {
    int margin;
    int y;

    /* Calculate a margin */
    margin = p->noise_width > CONST_DX ? p->noise_width : CONST_DX;

    /* Row by row */
    for(y = SKIP_DY; y < img->h - SKIP_DY; ++y)
        {
        int edge_left = y * img->w + margin; /* Skip left margin */
        int margin_right = y * img->w + img->w - (margin + 1);

        for(;;)
            {
            int edge_right;

            /* Search next edge position */
            edge_left = search_edge(
                img, edge_left, margin_right, edge_l, p->edge_c);

            /* End the loop if right margin reached */
            if(edge_left == margin_right)
                break;

            /* Search for nearby edge (maybe it's a thin line?) */
            edge_right = search_right(img,
                edge_left, margin_right, edge_l, p->edge_c, p->line_width);

            if(!show)
                /* Remove a noise */
                edge_left = clean_part(p->tmp, img, edge_left,
                    edge_right, edge_l, p->edge_c, p->noise_width);
            else
                {
                /* Just mark the edges */
                p->tmp[edge_left] = SHOW_YY;
                p->tmp[edge_right] = SHOW_YY;

                /* Skip processed row part */
                edge_left = edge_right + p->noise_width + 1;
                }
            }
        }

    /* Write cleaned luma plane */
    memcpy(img->yy, p->tmp, img->plane_sz);
    }

/* Search leftmost edge of all planes */
static int search_edge(
    vi3x_image_t const *img, int from, int to, int edge_l, int edge_c)
    {
    int e_cb, e_cr, e_yy;

    /* Luma plane */
    if(edge_l < VI3X_MAX)
        e_yy = search_single_left(img->yy, from, to, edge_l);
    else
        e_yy = to;

    /* Chroma planes */
    if(edge_c < VI3X_MAX)
        {
        e_cb = search_single_left(img->cb, from, to, edge_c);
        e_cr = search_single_left(img->cr, from, to, edge_c);
        }
    else
        {
        e_cb = to;
        e_cr = to;
        }

    /* Return leftmost (minimum) value */
    return e_yy < e_cb
        ? (e_yy < e_cr ? e_yy : e_cr) : (e_cb < e_cr ? e_cb : e_cr);
    }

/* Search rightmost (limited by line_width) edge of all planes */
static int search_right(vi3x_image_t const *img,
    int from, int to, int edge_l, int edge_c, int line_width)
    {
    int e_cb, e_cr, e_yy;

    if(from >= to)
        return to;

    /* Luma plane */
    if(edge_l < VI3X_MAX)
        e_yy = search_single_right(img->yy, from, to, edge_l, line_width);
    else
        e_yy = from + 1;

    /* Chroma planes */
    if(edge_c < VI3X_MAX)
        {
        e_cb = search_single_right(img->cb, from, to, edge_c, line_width);
        e_cr = search_single_right(img->cr, from, to, edge_c, line_width);
        }
    else
        {
        e_cb = from + 1;
        e_cr = from + 1;
        }

    /* Return rightmost (maximum) value */
    return e_yy > e_cb
        ? (e_yy > e_cr ? e_yy : e_cr) : (e_cb > e_cr ? e_cb : e_cr);
    }

/* Search edge by pixel values difference */
static int search_single_left(
    vi3x_sample_t const *px, int from, int to, int edge_level)
    {
    int x;
    int ad_prev = 0;

    from -= EDGE_DX;
    to -= EDGE_DX;

    for(x = from; x < to; ++x)
        {
        /* Get difference */
        int ad = abs_diff(px[x], px[x+EDGE_STEP]);

        /* Check difference on local maximum */
        if(ad < ad_prev && ad_prev > edge_level)
            return x - 1 + EDGE_DX; /* Edge found */

        /* Remember value */
        ad_prev = ad;
        }

    /* No luck */
    return to + EDGE_DX;
    }

/* Search nearby edge */
static int search_single_right(vi3x_sample_t const *px,
    int from, int to, int edge_level, int line_width)
    {
    int end;
    int x;

    /* Calculate right limit */
    end = from + line_width;
    if(end > to)
        end = to;

    /* Search flat area */
    for(x = from + 1; x < end; ++x)
        {
        /* Get differrence */
        int ad = abs_diff(px[x-EDGE_DX], px[x+EDGE_DX+1]);

        /* Check flatness */
        if(ad < edge_level)
            break;
        }

    /* Search next edge after flat area */
    x = search_single_left(px, x, end, edge_level);

    /* No luck */
    if(x == end)
        return from + 1;

    /* Got it */
    return x + 1;
    }

static inline int abs_diff(int a, int b)
    {
    return abs(a - b);
    }

/* Clean part of row around edges */
static int clean_part(vi3x_sample_t *dst, vi3x_image_t const *img,
    int l, int r, int edge_l, int edge_c, int noise_width)

    {
    int xy;

    /* Process left side */
    for(xy = l - noise_width; xy < l; ++xy)
        cond_avg(dst, img, xy, edge_l, edge_c);

    /* Process area between edges */
    for(; xy <= r; ++xy)
        edge_avg(dst, img, xy, edge_l, edge_c);

    /* Process right side */
    for(; xy <= r + noise_width; ++xy)
        cond_avg(dst, img, xy, edge_l, edge_c);

    /* Return end of processed part */
    return xy;
    }

/* Noisy area average with edge check */
static inline void cond_avg(vi3x_sample_t *dst,
    vi3x_image_t const *img, int xy, int edge_l, int edge_c)
    {
    /* Check for vertical pixels match */
    if(pix_match(img, xy, xy - img->w, edge_l, edge_c)
        && pix_match(img, xy, xy + img->w, edge_l, edge_c))
        {
        /* Check for horizontal pixels match */
        if(pix_match(img, xy, xy - 1, edge_l, edge_c)
            && pix_match(img, xy, xy + 1, edge_l, edge_c))
            /* Use 5 pixels */
            dst[xy] = (vi3x_sample_t)avg_5(img->yy + xy, img->w);
        else
            /* Use 3 pixels */
            dst[xy] = (vi3x_sample_t)avg_3(img->yy + xy, img->w);
        }
    }

/* Edge area average */
static inline void edge_avg(vi3x_sample_t *dst,
    vi3x_image_t const *img, int xy, int edge_l, int edge_c)
    {
    int sum;
    int n;

    /* Check for vertical pixels match */
    if(pix_match(img, xy, xy - img->w, edge_l, edge_c)
        && pix_match(img, xy, xy + img->w, edge_l, edge_c))
        {
        /* Top + bottom + center */
        sum = (int)(img->yy[xy]) * 2 + (int)(img->yy[xy-img->w])
            + (int)(img->yy[xy+img->w]);
        n = 4;

        /* Add left pixel */
        if(pix_match(img, xy, xy - 1, edge_l, edge_c))
            {
            sum += (int)(img->yy[xy]) + (int)(img->yy[xy-1]);
            n += 2;
            }

        /* Add right pixel */
        if(pix_match(img, xy, xy + 1, edge_l, edge_c))
            {
            sum += (int)(img->yy[xy]) + (int)(img->yy[xy+1]);
            n += 2;
            }

        /* Average matching pixels */
        dst[xy] = (vi3x_sample_t)((sum + n / 2) / n);
        }
    }

static inline int avg_3(vi3x_sample_t const *px, int w)
    {
    return ((int)(px[0]) * 2 + (int)(px[-w]) + (int)(px[w]) + 2) / 4;
    }

static inline int avg_5(vi3x_sample_t const *px, int w)
    {
    return ((int)(px[0]) * 4 + (int)(px[-1]) + (int)(px[1])
        + (int)(px[-w]) + (int)(px[w]) + 4) / 8;
    }

static inline int pix_match(
    vi3x_image_t const *img, int xy1, int xy2, int el, int ec)
    {
    return abs_diff(img->yy[xy1], img->yy[xy2]) <= el
        && abs_diff(img->cb[xy1], img->cb[xy2]) <= ec
        && abs_diff(img->cr[xy1], img->cr[xy2]) <= ec;
    }

/* Module help */
static void m_help(vi3x_help_t *h)
    {
    h->help_module(Tag, VI3X_VER_STR, "Vadim Druzhin",
        "Cartoon-oriented edge denoise filter",
        "This filter performs spatial denoise around edges\n"
        "in the luma plane. It is suitable for removing a 'dot crawl'\n"
        "(checkerboard-like noise caused by composite video processing).",
        ArgMap);
    }

static void help_denoise(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Number of denoise passes",
        "The bigger values will cause slower processing.",
        arg->def, 1, INT_MAX
        );
    }

static void help_edge_l(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Initial luma edge level (the pixel values difference)",
        "",
        arg->def, VI3X_MIN, VI3X_MAX
        );
    }

static void help_edge_c(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Chroma edge level (the pixel values difference)",
        "",
        arg->def, VI3X_MIN, VI3X_MAX
        );
    }

static void help_noise_width(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Width of the noisy area around edges (in pixels)",
        "A denoise will be performed in this area.",
        arg->def, 0, INT_MAX
        );
    }

static void help_line_width(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Maximum thickness of the vertical lines (in pixels)",
        "",
        arg->def, 0, INT_MAX
        );
    }

static void help_const_l(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Don't alter edge_l during processing",
        "By default edge_l is divided by 2 on each iteration,\n"
        "until it reaches value of edge_c * 2.",
        arg->def,
        "- Lower edge_l on each iteration",
        "- Use constant edge_l"
        );
    }

static void help_show(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Show the edges instead of performing a denoise", "",
        arg->def,
        "- Perform a denoise",
        "- Mark the edges"
        );
    }

