/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011-2012 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stddef.h>
#include <string.h>
#include <limits.h>
#include "vi3x_modules.h"

static char const *Tag="420p";

/* Instance local data */
typedef struct m_priv_s
    {
    int w;
    int h;
    int il;
    int mpeg1;
    int fps;
    int fps_div;
    vi3x_420_t i420;
    } m_priv_t;

static int init_420p(vi3x_module_t *m);
static int fini_420p(vi3x_module_t *m);
static int image_420p(vi3x_module_t *m, vi3x_image_t *image);
static void help_420p(vi3x_help_t *h);
static void help_w(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_h(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_il(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_mpeg1(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_fps(vi3x_arg_t const *arg, vi3x_help_t const *h);
static void help_fps_div(vi3x_arg_t const *arg, vi3x_help_t const *h);

/* Module arguments */
static vi3x_arg_t ArgMap[]=
    {
    { "w",       0, offsetof(m_priv_t, w), help_w },
    { "h",       0, offsetof(m_priv_t, h), help_h },
    { "il",      1, offsetof(m_priv_t, il), help_il },
    { "mpeg1",   0, offsetof(m_priv_t, mpeg1), help_mpeg1 },
    { "fps",    25, offsetof(m_priv_t, fps), help_fps },
    { "fps_div", 1, offsetof(m_priv_t, fps_div), help_fps_div },
    { NULL, 0, 0, NULL }
    };

/* Entry point */
VI3X_MAIN()
    {
    vi3x_reg_import(Tag, init_420p, sizeof(m_priv_t), help_420p);
    }

/* Module help */
static void help_420p(vi3x_help_t *h)
    {
    h->help_module(Tag, VI3X_VER_STR, "Vadim Druzhin",
        "Import raw frames in the planar 420p color space",
        "The frames are read from the standard input.",
        ArgMap);
    }

static void help_w(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Frame width",
        "The frame width must be even.",
        arg->def, 2, INT_MAX
        );
    }

static void help_h(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Frame height",
        "The frame height must be even, and must be\n"
        "divisible by 4 for interlaced frames.",
        arg->def, 2, INT_MAX
        );
    }

static void help_il(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Interlacing mode of chroma planes", "",
        arg->def,
        "- use progressive upsampling;",
        "- use interlaced upsampling."
        );
    }

static void help_mpeg1(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_bool(arg->name,
        "Sample position of chroma planes", "",
        arg->def,
        "- assume MPEG-2 sample position;",
        "- assume MPEG-1 sample position."
        );
    }

static void help_fps(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Frame rate", "",
        arg->def, 1, INT_MAX
        );
    }

static void help_fps_div(vi3x_arg_t const *arg, vi3x_help_t const *h)
    {
    h->help_param_range(arg->name,
        "Frame rate divisor",
        "The divisor required to specify non-integer frame rates.\n"
        "I.e., NTSC frame rate is: fps=30000,fps_div=1001.",
        arg->def, 1, INT_MAX
        );
    }

/* Instance initialization */
static int init_420p(vi3x_module_t *m)
    {
    /* Parse and verify arguments */
    if(!m->vi3x.parse_args(m->priv, m->args, ArgMap, Tag))
        return VI3X_RET_ERR;

    if(m->priv->w <= 0 || m->priv->h <= 0)
        {
        m->vi3x.err("[%s] Error: Incorrect frame size: %d x %d !\n",
            Tag, m->priv->w, m->priv->h);
        return VI3X_RET_ERR;
        }

    if((m->priv->w & 1) || (m->priv->h & 1))
        {
        m->vi3x.err("[%s] Error: Uneven frame size: %d x %d!\n",
            Tag, m->priv->w, m->priv->h);
        return VI3X_RET_ERR;
        }

    if(m->priv->il && (m->priv->h & 3))
        {
        m->vi3x.err("[%s] Error: Height of interlaced frame "
            "must be multiple of 4: height=%d !\n", Tag, m->priv->h);
        return VI3X_RET_ERR;
        }

    if(m->priv->fps <= 0)
        {
        m->vi3x.err("[%s] Error: Incorrect FPS: %d !\n", Tag, m->priv->fps);
        return VI3X_RET_ERR;
        }

    if(m->priv->fps_div <= 0)
        {
        m->vi3x.err("[%s] Error: Incorrect FPS_DIV: %d !\n", Tag, m->priv->fps_div);
        return VI3X_RET_ERR;
        }

    /* Initialize input image buffer */
    m->vi3x.init_420(&m->priv->i420, m->priv->w, m->priv->h, NULL);
    m->vi3x.init_420(&m->priv->i420, m->priv->w, m->priv->h,
        m->vi3x.alloc(m->priv->i420.img_sz, Tag));
    m->priv->i420.il = m->priv->il;

    /* Publish module interface */
    m->w = m->priv->w;
    m->h = m->priv->h;
    m->fps = m->priv->fps;
    m->fps_div = m->priv->fps_div;
    m->fini = fini_420p;
    m->process_image = image_420p;

    return VI3X_RET_OK;
    }

/* Instance finalization */
static int fini_420p(vi3x_module_t *m)
    {
    /* Free instance memory */
    m->vi3x.free(m->priv->i420.yy);

    return VI3X_RET_OK;
    }

/* Frame parser */
static int image_420p(vi3x_module_t *m, vi3x_image_t *image)
    {
    int ret;

    /* Read whole frame, color planes order assumed as in default vi3x_420_t */
    ret=m->vi3x.read(m->priv->i420.yy, m->priv->i420.img_sz);
    if(ret < m->priv->i420.img_sz)
        {
        if(0 != ret)
            m->vi3x.warn("[%s] Warning: Frame %ld - partial frame read!\n",
                Tag, m->frame_n);
        return VI3X_RET_FINISH;
        }

    /* Upsample frame for future processing */
    if(m->priv->mpeg1)
        m->vi3x.upsample_mpeg1(image, &(m->priv->i420));
    else
        m->vi3x.upsample_mpeg2(image, &(m->priv->i420));

    return m->out_image(m, image);
    }
