/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include "vi3x_types.h"
#include "vi3x_image.h"

#define MAX_DELTA ((VI3X_MAX - VI3X_MIN) / 4)

static void upsample_rows_mpeg1(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h);
static void upsample_rows_mpeg2(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h);
static void upsample_frame(vi3x_sample_t *dst, int w, int h);
static void upsample_fields(vi3x_sample_t *dst, int w, int h);
static inline int delta12(int p0, int p1, int p2);
static inline int delta34(int p0, int p1, int p2);
static inline int delta_cr(int p0, int p1, int p2, int p3);
static inline void inter12(
    vi3x_sample_t *dst1, vi3x_sample_t *dst2, int s0, int s1, int s2);
static inline void inter34(
    vi3x_sample_t *dst1, vi3x_sample_t *dst2, int s0, int s1, int s2);
static inline void inter_cr(
    vi3x_sample_t *dst1, vi3x_sample_t *dst2, int s0, int s1, int s2, int s3);
static void downsample_mpeg2_frame(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h);
static void downsample_mpeg2_fields(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h);
static void downsample_mpeg1_frame(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h);
static void downsample_mpeg1_fields(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h);
static void resample_fields_to_frame(vi3x_sample_t *buf, int w, int h);
static void resample_frame_to_fields(vi3x_sample_t *buf, int w, int h);

vi3x_sample_t *vi3x_init_420(
    vi3x_420_t *image, int w, int h, vi3x_sample_t *buf)
    {
    image->w=w;
    image->h=h;
    image->yy_count=w * h;
    image->yy_sz=image->yy_count * sizeof(*image->yy);
    image->cw=(w + 1) / 2;
    image->ch=(h + 1) / 2;
    image->cc_count=image->cw * image->ch;
    image->cc_sz=image->cc_count * sizeof(*image->yy);
    image->img_sz=image->yy_sz + image->cc_sz * 2;
    image->yy=buf;
    if(NULL != buf)
        buf+=image->yy_count;
    image->cb=buf;
    if(NULL != buf)
        buf+=image->cc_count;
    image->cr=buf;
    if(NULL != buf)
        buf+=image->cc_count;
    image->il=0;

    return buf;
    }

vi3x_sample_t *vi3x_init_image(
    vi3x_image_t *image, int w, int h, vi3x_sample_t *buf)
    {
    image->w=w;
    image->h=h;
    image->count=w * h;
    image->plane_sz=image->count * sizeof(*image->yy);
    image->img_sz=image->plane_sz * 3;
    image->yy=buf;
    if(NULL != buf)
        buf+=image->count;
    image->cb=buf;
    if(NULL != buf)
        buf+=image->count;
    image->cr=buf;
    if(NULL != buf)
        buf+=image->count;
    image->il=0;

    return buf;
    }

/* Image width must be multiple of 2 and at least 4.
   Image height must bet multiple of 2 and at least 4
   for progressive frame.
   Image height must bet multiple of 4 and at least 8
   for interlaced frame. */
void vi3x_upsample_mpeg1(vi3x_image_t *dst, vi3x_420_t const *src)
    {
    /* Copy luma plane */
    memcpy(dst->yy, src->yy, dst->plane_sz);

    /* Upsample blue chroma plane */
    upsample_rows_mpeg1(dst->cb, src->cb, src->cw, src->ch);
    if(src->il)
        upsample_fields(dst->cb, dst->w, dst->h);
    else
        upsample_frame(dst->cb, dst->w, dst->h);

    /* Upsample red chroma plane */
    upsample_rows_mpeg1(dst->cr, src->cr, src->cw, src->ch);
    if(src->il)
        upsample_fields(dst->cr, dst->w, dst->h);
    else
        upsample_frame(dst->cr, dst->w, dst->h);

    dst->il=src->il;
    }

/* Image width must be multiple of 2 and at least 6.
   Image height must bet multiple of 2 and at least 4
   for progressive frame.
   Image height must bet multiple of 4 and at least 8
   for interlaced frame. */
void vi3x_upsample_mpeg2(vi3x_image_t *dst, vi3x_420_t const *src)
    {
    /* Copy luma plane */
    memcpy(dst->yy, src->yy, dst->plane_sz);

    /* Upsample blue chroma plane */
    upsample_rows_mpeg2(dst->cb, src->cb, src->cw, src->ch);
    if(src->il)
        upsample_fields(dst->cb, dst->w, dst->h);
    else
        upsample_frame(dst->cb, dst->w, dst->h);

    /* Upsample red chroma plane */
    upsample_rows_mpeg2(dst->cr, src->cr, src->cw, src->ch);
    if(src->il)
        upsample_fields(dst->cr, dst->w, dst->h);
    else
        upsample_frame(dst->cr, dst->w, dst->h);

    dst->il=src->il;
    }

/* Image width must be multiple of 2.
   Image height must be multiple of 2
   for progressive frame.
   Image height must be multiple of 4
   for interlaced frame. */
void vi3x_downsample_mpeg1(vi3x_420_t *dst, vi3x_image_t const *src)
    {
    /* Copy luma plane */
    memcpy(dst->yy, src->yy, dst->yy_sz);

    if(src->il)
        {
        /* Downsample blue chroma plane */
        downsample_mpeg1_fields(dst->cb, src->cb, src->w, src->h);
        /* Downsample red chroma plane */
        downsample_mpeg1_fields(dst->cr, src->cr, src->w, src->h);
        }
    else
        {
        /* Downsample blue chroma plane */
        downsample_mpeg1_frame(dst->cb, src->cb, src->w, src->h);
        /* Downsample red chroma plane */
        downsample_mpeg1_frame(dst->cr, src->cr, src->w, src->h);
        }

    dst->il=src->il;
    }

/* Image width must be multiple of 2.
   Image height must be multiple of 2
   for progressive frame.
   Image height must be multiple of 4
   for interlaced frame. */
void vi3x_downsample_mpeg2(vi3x_420_t *dst, vi3x_image_t const *src)
    {
    /* Copy luma plane */
    memcpy(dst->yy, src->yy, dst->yy_sz);

    if(src->il)
        {
        /* Downsample blue chroma plane */
        downsample_mpeg2_fields(dst->cb, src->cb, src->w, src->h);
        /* Downsample red chroma plane */
        downsample_mpeg2_fields(dst->cr, src->cr, src->w, src->h);
        }
    else
        {
        /* Downsample blue chroma plane */
        downsample_mpeg2_frame(dst->cb, src->cb, src->w, src->h);
        /* Downsample red chroma plane */
        downsample_mpeg2_frame(dst->cr, src->cr, src->w, src->h);
        }

    dst->il=src->il;
    }

/* Image height must be multiple of 4 and at least 4 */
void vi3x_fields_to_frame(vi3x_image_t *img)
    {
    /* Resample blue chroma plane */
    resample_fields_to_frame(img->cb, img->w, img->h);
    /* Resample red chroma plane */
    resample_fields_to_frame(img->cr, img->w, img->h);
    img->il=0;
    }

/* Image height must be multiple of 4 and at least 8 */
void vi3x_frame_to_fields(vi3x_image_t *img)
    {
    /* Resample blue chroma plane */
    resample_frame_to_fields(img->cb, img->w, img->h);
    /* Resample red chroma plane */
    resample_frame_to_fields(img->cr, img->w, img->h);
    img->il=1;
    }

/*
Upsample mpeg1 chroma rows:
~~~~~~~~~~~~~~~~~~~~~~~~~~~
S[ 0   1 ]
  +O+ +O+
  | | | |
  v v v v
  O O O O
  . . . .
D[0 1 2 3]

Using proportional upsampling:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
D[n*2]=S[n] + (S[n-2] - S[n+2]) / 8
D[n*2+1]=S[n] - (S[n-2] - S[n+2]) / 8
*/
/* Source width must be at least 2 */
static void upsample_rows_mpeg1(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h)
    {
    vi3x_sample_t *end;

    end=dst + w * h * 4;
    while(dst < end)
        {
        vi3x_sample_t const *rend=src + w; /* End of row */
        int s0=*src++; /* Source pixels */
        int s1=*src++;
        /* Write first pair of destination pixels */
        inter12(dst, dst + 1, s0, s0, s1);
        dst+=2;
        while(src < rend)
            {
            /* Write pair of destination pixels */
            inter12(dst, dst + 1, s0, s1, *src);
            dst+=2;
            /* Remember source pixels */
            s0=s1;
            s1=*src++;
            }
        /* Write last pair of destination pixels */
        inter12(dst, dst + 1, s0, s1, s1);
        dst+=2;
        /* Skip destination row */
        dst+=w * 2;
        }
    }

/*
Upsample mpeg2 chroma rows:
~~~~~~~~~~~~~~~~~~~~~~~~~~~
S[0   1  ]
  O-+ O-+
  | | | |
  v v v v
  O O O O
  . . . .
D[0 1 2 3]

Using Catmull-Rom interpolation:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
D[n*2]=S[n]
D[n*2+1]=(9 * (S[n] + S[n+2]) - (S[n-2] + S[n+4])) / 16
*/
/* Source width must be at least 3 */
static void upsample_rows_mpeg2(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h)
    {
    vi3x_sample_t *end;

    end=dst + w * h * 4;
    while(dst < end)
        {
        vi3x_sample_t const *rend=src + w; /* End of row */
        int s0=*src++; /* Source pixels */
        int s1=*src++;
        int s2=*src++;
        /* Write first pair of destination pixels */
        inter_cr(dst, dst + 1, s0, s0, s1, s2);
        dst+=2;
        while(src < rend)
            {
            /* Write pair of destination pixels */
            inter_cr(dst, dst + 1, s0, s1, s2, *src);
            dst+=2;
            /* Remember source pixels */
            s0=s1;
            s1=s2;
            s2=*src++;
            }
        /* Write last pairs of destination pixels */
        inter_cr(dst, dst + 1, s0, s1, s2, s2);
        dst+=2;
        inter_cr(dst, dst + 1, s1, s2, s2, s2);
        dst+=2;
        /* Skip destination row */
        dst+=w*2;
        }
    }

/*
Upsample chroma frame:
~~~~~~~~~~~~~~~~~~~~~~
S[0] O +-> O D[0]
     . +-> O D[1]
S[2] O +-> O D[2]
     . +-> O D[3]

Using proportional upsampling:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
D[n]=S[n] + (S[n-2] - S[n+2]) / 8
D[n+1]=S[n] - (S[n-2] - S[n+2]) / 8
*/
/* Buffer height must be multiple of 2 and at least 4 */
static void upsample_frame(vi3x_sample_t *dst, int w, int h)
    {
    vi3x_sample_t *end;
    int x;

    end=dst + w * (h - 2);
    for(x=0; x < w; ++x)
        {
        vi3x_sample_t *dx=dst + x; /* Current column */
        int s0=dx[0]; /* Source pixels */
        int s1=dx[w*2];
        /* Write first pair of destination pixels */
        inter12(dx, dx + w, s0, s0, s1);
        dx+=w * 2;
        while(dx < end)
            {
            /* Write pair of destination pixels */
            inter12(dx, dx + w, s0, s1, dx[w*2]);
            s0=s1; /* Remember source pixels */
            dx+=w * 2;
            s1=dx[0];
            }
        /* Write last pair of destination pixels */
        inter12(dx, dx + w, s0, s1, s1);
        }
    }

/*
Upsample interlaced chroma fields:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
S[0] O +-> O D[0]
     . |                 +-> O D[1]
       +-> O D[2] S[2] O |
                       . +-> O D[3]
S[4] O +-> O D[4]
     . |                 +-> O D[5]
       +-> O D[6] S[6] O |
                       . +-> O D[7]

Using proportional upsampling:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
D[n]=S[n] + (S[n-4] - S[n+4]) / 16
D[n+1]=S[n+2] + 3 * (S[n-2] - S[n+6]) / 16
D[n+2]=S[n] - 3 * (S[n-4] - S[n+4]) / 16
D[n+3]=S[n+2] - (S[n-2] - S[n+6]) / 16
*/
/* Buffer height must be multiple of 4 and at least 8 */
static void upsample_fields(vi3x_sample_t *dst, int w, int h)
    {
    vi3x_sample_t *end;
    int x;

    end=dst + w * (h - 4);
    for(x=0; x < w; ++x)
        {
        vi3x_sample_t *dx=dst + x; /* Current column */
        /* Source pixels */
        int st0=dx[0];
        int sb0=dx[w*2];
        int st1=dx[w*4];
        int sb1=dx[w*6];
        /* Write even destination pixels */
        inter34(dx, dx + w * 2, st0, st0, st1);
        /* Write odd destination pixels */
        inter34(dx + w * 3, dx + w, sb1, sb0, sb0);
        dx+=w * 4;
        while(dx < end)
            {
            /* Read source pixels */
            st1=dx[0];
            sb1=dx[w*2];
            /* Write even destination pixels */
            inter34(dx, dx + w * 2, st0, st1, dx[w*4]);
            /* Remember source pixel */
            st0=st1;
            /* Write odd destination pixels */
            inter34(dx + w * 3, dx + w, dx[w*6], sb1, sb0);
            /* Remember source pixel */
            sb0=sb1;
            dx+=w * 4;
            }
        /* Read source pixels */
        st1=dx[0];
        sb1=dx[w*2];
        /* Write even destination pixels */
        inter34(dx, dx + w * 2, st0, st1, st1);
        /* Write odd destination pixels */
        inter34(dx + w * 3, dx + w, sb1, sb1, sb0);
        }
    }

static inline int delta12(int p0, int p1, int p2)
    {
    /* Test for edges */
    if(abs(p0 - p1) > MAX_DELTA)
        p0=p1;
    if(abs(p2 - p1) > MAX_DELTA)
        p2=p1;

    return (p0 - p2 + 4) / 8;
    }

static inline int delta34(int p0, int p1, int p2)
    {
    /* Test for edges */
    if(abs(p0 - p1) > MAX_DELTA)
        p0=p1;
    if(abs(p2 - p1) > MAX_DELTA)
        p2=p1;

    return (p0 - p2 + 8) / 16;
    }

static inline int delta_cr(int p0, int p1, int p2, int p3)
    {
    /* Test for edges */
    if(abs(p0 - p1) > MAX_DELTA)
        p0=p1;
    if(abs(p2 - p1) > MAX_DELTA)
        p2=p1;
    if(abs(p3 - p2) > MAX_DELTA)
        p3=p2;

    return (p2 * 9 - p1 * 7 - p0 - p3 + 8) / 16;
    }

static inline void inter12(
    vi3x_sample_t *dst1, vi3x_sample_t *dst2, int s0, int s1, int s2)
    {
    int d1, d2;

    /* Interpolate */
    d1=delta12(s0, s1, s2);
    d2=s1 - d1;
    d1+=s1;

    /* Fix overflow */
    if(d1 < VI3X_MIN)
        {
        d1=VI3X_MIN;
        d2=s1 * 2; /* VI3X_MIN assumed to be 0*/
        }
    else if(d1 > VI3X_MAX)
        {
        d1=VI3X_MAX;
        d2=s1 * 2 - VI3X_MAX;
        }
    else if(d2 < VI3X_MIN)
        {
        d2=VI3X_MIN;
        d1=s1 * 2; /* VI3X_MIN assumed to be 0 */
        }
    else if(d2 > VI3X_MAX)
        {
        d2=VI3X_MAX;
        d1=s1 * 2 - VI3X_MAX;
        }

    /* Store pixels */
    *dst1=(vi3x_sample_t)d1;
    *dst2=(vi3x_sample_t)d2;
    }

static inline void inter34(
    vi3x_sample_t *dst1, vi3x_sample_t *dst2, int s0, int s1, int s2)
    {
    int d1, d2;

    /* Interpolate */
    d1=delta34(s0, s1, s2);
    d2=s1 - d1 * 3;
    d1+=s1;

    /* Fix overflow */
    if(d1 < VI3X_MIN)
        {
        d1=VI3X_MIN;
        d2=s1 * 4; /* VI3X_MIN assumed to be 0 */
        }
    else if(d1 > VI3X_MAX)
        {
        d1=VI3X_MAX;
        d2=s1 * 4 - VI3X_MAX * 3;
        }
    else if(d2 < VI3X_MIN)
        {
        d1-=(-d2 + 2) / 3; /* VI3X_MIN assumed to be 0 */
        d2=s1 * 4 - d1 * 3;
        }
    else if(d2 > VI3X_MAX)
        {
        d1+=(d2 - VI3X_MAX + 2) / 3;
        d2=s1 * 4 - d1 * 3;
        }

    /* Store pixels */
    *dst1=(vi3x_sample_t)d1;
    *dst2=(vi3x_sample_t)d2;
    }

static inline void inter_cr(
    vi3x_sample_t *dst1, vi3x_sample_t *dst2, int s0, int s1, int s2, int s3)
    {
    int d;

    /* Interpolate */
    d=s1 + delta_cr(s0, s1, s2, s3);

    /* Fix overflow */
    if(d < VI3X_MIN)
        d=VI3X_MIN;
    else if(d > VI3X_MAX)
        d=VI3X_MAX;

    /* Store pixels */
    *dst1=(vi3x_sample_t)s1;
    *dst2=(vi3x_sample_t)d;
    }

/*
Downsample chroma using mpeg2 sample position:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 S[0,0] S[0,1] S[0,2] S[0,3]
 D[0,0]        D[0,1]
 S[1,0] S[1,1] S[1,2] S[1,3]

 S[2,0] S[2,1] S[2,2] S[2,3]
 D[1,0]        D[1,1]
 S[3,0] S[3,1] S[3,2] S[3,3]

D[m,n]=(S[m*2,n*2] + S[m*2+1,n*2]) / 2
*/
/* Source height must be multiple of 2,
   source width must be multiple of 2 */
static void downsample_mpeg2_frame(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h)
    {
    vi3x_sample_t const *end;

    for(end=src + w * h; src < end; src+=w)
        {
        vi3x_sample_t const *rend;
        for(rend=src + w; src < rend; src+=2)
            {
            *dst=(vi3x_sample_t)(((int)(src[0]) + (int)(src[w]) + 1) / 2);
            ++dst;
            }
        }
    }

/*
Downsample chroma using mpeg2 interlaced sample position:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 S[0,0] S[0,1] S[0,2] S[0,3]
 D[0,0]        D[0,1]
                             S[1,0] S[1,1] S[1,2] S[1,3]

 S[2,0] S[2,1] S[2,2] S[2,3]
                             D[1,0]        D[1,1]
                             S[3,0] S[3,1] S[3,2] S[3,3]

De[m,n]=(3 * S[m*2,n*2] + S[m*2+2,n*2]) / 4
Do[m,n]=(S[m*2-1,n*2] + 3 * S[m*2+1,n*2]) / 4
*/
/* Source height must be multiple of 4,
   source width must be multiple of 2 */
static void downsample_mpeg2_fields(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h)
    {
    vi3x_sample_t const *end;

    for(end=src + w * h; src < end; src+=w * 2)
        {
        vi3x_sample_t const *rend;
        /* Even rows */
        for(rend=src + w; src < rend; src+=2)
            {
            *dst=(vi3x_sample_t)(((int)(src[0]) * 3 + (int)(src[w*2]) + 2) / 4);
            ++dst;
            }
        /* Odd rows */
        for(rend=src + w; src < rend; src+=2)
            {
            *dst=(vi3x_sample_t)(((int)(src[0]) + (int)(src[w*2]) * 3 + 2) / 4);
            ++dst;
            }
        }
    }

/*
Downsample chroma using mpeg1 sample position:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 S[0,0] X S[0,1] S[0,2] X S[0,3]
        D[0,0]          D[0,1]
 S[1,0] X S[1,1] S[1,2] X S[1,3]

 S[2,0] X S[2,1] S[2,2] X S[2,3]
        D[1,0]          D[1,1]
 S[3,0] X S[3,1] S[3,2] X S[3,3]

D[m,n]=(S[m*2,n*2] + S[m*2+1,n*2] + S[m*2,n*2+1] + S[m*2+1,n*2+1]) / 4
*/
/* Source height must be multiple of 2,
   source width must be multiple of 2 */
static void downsample_mpeg1_frame(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h)
    {
    vi3x_sample_t const *end;

    for(end=src + w * h; src < end; src+=w)
        {
        vi3x_sample_t const *rend;
        for(rend=src + w; src < rend; src+=2)
            {
            *dst=(vi3x_sample_t)(((int)(src[0]) + (int)(src[1])
                    + (int)(src[w]) + (int)(src[w+1]) + 2) / 4);
            ++dst;
            }
        }
    }

/*
Downsample chroma using mpeg1 interlaced sample position:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 S[0,0] X S[0,1] S[0,2] X S[0,3]
        D[0,0]          D[0,1]
                                 S[1,0] X S[1,1] S[1,2] X S[1,3]

 S[2,0] X S[2,1] S[2,2] X S[2,3]
                                        D[1,0]          D[1,1]
                                 S[3,0] X S[3,1] S[3,2] X S[3,3]

De[m,n]=(3 * (S[m*2,n*2] + S[m*2,n*2+1]) + S[m*2+2,n*2] + S[m*2+2,n*2+1]) / 8
Do[m,n]=(S[m*2-1,n*2] + S[m*2-1,n*2+1] + 3 * (S[m*2+1,n*2] + S[m*2+1,n*2+1])) / 8
*/
/* Source height must be multiple of 4,
   source width must be multiple of 2 */
static void downsample_mpeg1_fields(
    vi3x_sample_t *dst, vi3x_sample_t const *src, int w, int h)
    {
    vi3x_sample_t const *end;

    for(end=src + w * h; src < end; src+=w*2)
        {
        vi3x_sample_t const *rend;
        /* Even rows */
        for(rend=src + w; src < rend; src+=2)
            {
            *dst=(vi3x_sample_t)((((int)(src[0]) + (int)(src[1])) * 3
                    + (int)(src[w*2]) + (int)(src[w*2+1]) + 4) / 8);
            ++dst;
            }
        /* Odd rows */
        for(rend=src + w; src < rend; src+=2)
            {
            *dst=(vi3x_sample_t)(((int)(src[0]) + (int)(src[1])
                    + ((int)(src[w*2]) + (int)(src[w*2+1])) * 3 + 4) / 8);
            ++dst;
            }
        }
    }

void vi3x_copy_image(vi3x_image_t *dst, vi3x_image_t const *src)
    {
    memcpy(dst->yy, src->yy, dst->plane_sz);
    memcpy(dst->cb, src->cb, dst->plane_sz);
    memcpy(dst->cr, src->cr, dst->plane_sz);
    dst->il=src->il;
    }

/* Buffer height must be multiple of 4 and at least 4 */
static void resample_fields_to_frame(vi3x_sample_t *dst, int w, int h)
    {
    vi3x_sample_t *end;
    int x;

    end=dst + w * h;
    for(x=0; x < w; ++x)
        {
        vi3x_sample_t *dx=dst + x; /* Current column */
        /* Reconstruct source pixels */
        int s0=((int)(dx[0]) * 3 + (int)(dx[w*2]) + 2) / 4;
        int s1=((int)(dx[w]) + (int)(dx[w*3]) * 3 + 2) / 4;
        /* Write destination pixels */
        inter12(dx, dx + w, s0, s0, s1);
        dx+=w * 4;
        while(dx < end)
            {
            /* Reconstruct source pixels */
            int s2=((int)(dx[0]) * 3 + (int)(dx[w*2]) + 2) / 4;
            int s3=((int)(dx[w]) + (int)(dx[w*3]) * 3 + 2) / 4;
            /* Write destination pixels */
            inter12(dx - w * 2, dx - w, s0, s1, s2);
            s0=s2; /* Remember source pixel */
            inter12(dx, dx + w, s1, s2, s3);
            s1=s3; /* Remember source pixel */
            dx+=w * 4;
            }
        /* Write destination pixels */
        inter12(dx - w * 2, dx - w, s0, s1, s1);
        }
    }

static inline void avg4r(int *a, int *b, vi3x_sample_t const *src, int w)
    {
    *a=((int)(src[0]) + (int)(src[w]) + 1) / 2;
    *b=((int)(src[w*2]) + (int)(src[w*3]) + 1) / 2;
    }

/* Buffer height must be multiple of 4 and at least 8 */
static void resample_frame_to_fields(vi3x_sample_t *dst, int w, int h)
    {
    vi3x_sample_t *end;
    int x;

    end=dst + w * (h - 4);
    for(x=0; x < w; ++x)
        {
        vi3x_sample_t *dx=dst + x; /* Current column */
        int st0, st1, sb0, sb1;
        /* Reconstruct source pixels */
        avg4r(&st0, &sb0, dx, w);
        avg4r(&st1, &sb1, dx + w * 4, w);
        /* Write even destination pixels */
        inter34(dx, dx + w * 2, st0, st0, st1);
        /* Write odd destination pixels */
        inter34(dx + w * 3, dx + w, sb1, sb0, sb0);
        dx+=w * 4;
        while(dx < end)
            {
            int st2, sb2;
            /* Reconstruct source pixels */
            avg4r(&st2, &sb2, dx + w * 4, w);
            /* Write even destination pixels */
            inter34(dx, dx + w * 2, st0, st1, st2);
            /* Remember source pixels */
            st0=st1;
            st1=st2;
            /* Write odd destination pixels */
            inter34(dx + w * 3, dx + w, sb2, sb1, sb0);
            /* Remember source pixels */
            sb0=sb1,
            sb1=sb2;
            dx+=w * 4;
            }

        /* Write even destination pixels */
        inter34(dx, dx + w * 2, st0, st1, st1);
        /* Write odd destination pixels */
        inter34(dx + w * 3, dx + w, sb1, sb1, sb0);
        }
    }
