/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011-2012 Vadim Druzhin <cdslow@mail.ru>
*/

#include <stdlib.h>
#include <string.h>
#include "vi3x_io.h"
#include "vi3x_utils.h"
#include "vi3x_help.h"
#include "vi3x_image.h"
#include "vi3x_modules.h"
#include "modules_main.h"

typedef struct reg_module_s
    {
    char name[VI3X_NAME_SZ];
    vi3x_module_f *init;
    vi3x_help_f *help;
    int priv_sz;
    struct reg_module_s *next;
    } reg_module_t;

typedef struct module_list_s
    {
    reg_module_t *first;
    } module_list_t;

static module_list_t list_import;
static module_list_t list_export;
static module_list_t list_filter;

static const vi3x_t Vi3x=
    {
    vi3x_init_420,
    vi3x_init_image,
    vi3x_copy_image,
    vi3x_upsample_mpeg1,
    vi3x_upsample_mpeg2,
    vi3x_downsample_mpeg1,
    vi3x_downsample_mpeg2,
    vi3x_fields_to_frame,
    vi3x_frame_to_fields,
    vi3x_read,
    vi3x_write,
    vi3x_err,
    vi3x_warn,
    vi3x_info,
    vi3x_info_va,
    vi3x_parse_args,
    vi3x_alloc,
    vi3x_free
    };

static void reg_module(module_list_t *list, char const *m_name,
    vi3x_module_f *m_init, int priv_sz, vi3x_help_f *m_help);
static vi3x_module_t *create_module(
    char const *m_name, module_list_t const *list, int with_range);
static reg_module_t const *search_module(
    char const *m_name, module_list_t const *list, char const *tag);
static int match_module(module_list_t const *list, vi3x_module_t const *m);
static void help_all(module_list_t const *list, int help_short);
static void vi3x_funcs(vi3x_t *vi3x);
static void help_module(
    char const *m_name, module_list_t const *list, char const *tag);
static void help_def(char const *m_name, vi3x_help_t const *hf);

void vi3x_register_modules(void)
    {
    unsigned i;

    for(i=0; i < sizeof(ModulesMain) / sizeof(*ModulesMain); ++i)
        ModulesMain[i]();
    }

void vi3x_reg_import(char const *m_name,
    vi3x_module_f *m_init, int priv_sz, vi3x_help_f *m_help)
    {
    reg_module(&list_import, m_name, m_init, priv_sz, m_help);
    }

void vi3x_reg_export(char const *m_name,
    vi3x_module_f *m_init, int priv_sz, vi3x_help_f *m_help)
    {
    reg_module(&list_export, m_name, m_init, priv_sz, m_help);
    }

void vi3x_reg_filter(char const *m_name,
    vi3x_module_f *m_init, int priv_sz, vi3x_help_f *m_help)
    {
    reg_module(&list_filter, m_name, m_init, priv_sz, m_help);
    }

static void reg_module(module_list_t *list, char const *m_name,
    vi3x_module_f *m_init, int priv_sz, vi3x_help_f *m_help)
    {
    reg_module_t *next;

    if(NULL == m_name)
        {
        vi3x_err("[reg_module] Error: NULL module name!\n");
        return;
        }

    if(strlen(m_name) >= VI3X_NAME_SZ)
        {
        vi3x_err("[reg_module] Error: Module name too long: '%s'!\n", m_name);
        return;
        }

    if(NULL == m_init)
        {
        vi3x_err("[reg_module] Error: NULL m_init in module '%s'!\n", m_name);
        return;
        }

    if(priv_sz < 0)
        {
        vi3x_err(
            "[reg_module] Error: Negative priv_sz in module '%s'!\n", m_name);
        return;
        }

    next = vi3x_alloc(sizeof(*next), "reg_module");

    strncpy(next->name, m_name, VI3X_NAME_SZ);
    next->init = m_init;
    next->priv_sz = priv_sz;
    next->help = m_help;
    next->next = NULL;

    if(NULL == list->first)
        /* Place first module into the list */
        list->first = next;
    else if(strcmp(next->name, list->first->name) < 0)
        {
        /* Insert new module before the first one */
        next->next = list->first;
        list->first = next;
        }
    else
        {
        reg_module_t *m;

        /* Search the insertion point */
        for(m = list->first; NULL != m->next; m = m->next)
            {
            if(strcmp(next->name, m->next->name) < 0)
                break;
            }

        /* Insert new module in alphabetical order */
        next->next = m->next;
        m->next = next;
        }
    }

vi3x_module_t *vi3x_create_import(char const *m_name)
    {
    vi3x_module_t *m;

    m=create_module(m_name, &list_import, 0);
    if(NULL == m)
        vi3x_err("[vi3x_create_import] Error: Unknown import module: '%s'!\n", m_name);

    return m;
    }

vi3x_module_t *vi3x_create_export(char const *m_name)
    {
    vi3x_module_t *m;

    m=create_module(m_name, &list_export, 0);
    if(NULL == m)
        vi3x_err("[vi3x_create_export] Error: Unknown export module: '%s'!\n", m_name);

    return m;
    }

vi3x_module_t *vi3x_create_filter(char const *m_name)
    {
    vi3x_module_t *m;

    m=create_module(m_name, &list_filter, 1);
    if(NULL == m)
        vi3x_err("[vi3x_create_filter] Error: Unknown filter module: '%s'!\n", m_name);

    return m;
    }

static vi3x_module_t *create_module(
    char const *m_name, module_list_t const *list, int with_range)
    {
    vi3x_module_t *m;
    char const *s;
    reg_module_t const *rm;
    int range_count;
    char const *tag="create_module";

    rm = search_module(m_name, list, tag);
    if(NULL == rm)
        return NULL;

    s = m_name + strlen(rm->name);
    if(VI3X_D_RANGE == *s)
        {
        if(!with_range)
            {
            vi3x_err("[%s] Error: Range can't be used here: '%s'!\n",
                tag, m_name);
            return NULL;
            }
        ++s;
        range_count = vi3x_count_ranges(s, tag);
        if(0 == range_count)
            return NULL;
        }
    else
        range_count = 1;

    m = vi3x_alloc(
        sizeof(*m) + sizeof(*m->range) * range_count + rm->priv_sz, tag);

    m->range = (long *)(m + 1);
    if(range_count > 1)
        {
        s = vi3x_parse_ranges(m->range, s, tag);
        if(NULL == s)
            {
            /* Must never hapen */
            vi3x_free(m);
            return NULL;
            }
        m->disabled = 1;
        }
    else
        {
        *m->range = -1;
        m->disabled = 0;
        }
    if(0 != *s)
        ++s;

    if(rm->priv_sz > 0)
        m->priv = (struct m_priv_s *)(m->range + range_count);

    m->name = rm->name;
    m->init = rm->init;
    m->args = s;
    m->frame_n = 0l;
    m->range_ptr = m->range;
    vi3x_funcs(&m->vi3x);

    return m;
    }

void vi3x_free_module(vi3x_module_t *m)
    {
    vi3x_free(m);
    }

void vi3x_help_import(char const *m_name)
    {
    help_module(m_name, &list_import, "vi3x_help_import");
    }

void vi3x_help_export(char const *m_name)
    {
    help_module(m_name, &list_export, "vi3x_help_export");
    }

void vi3x_help_filter(char const *m_name)
    {
    help_module(m_name, &list_filter, "vi3x_help_filter");
    }

static void help_module(
    char const *m_name, module_list_t const *list, char const *tag)
    {
    reg_module_t const *rm;
    vi3x_help_t hf;

    rm=search_module(m_name, list, tag);
    if(NULL == rm)
        {
        vi3x_err("[%s] Error: Unknown module: '%s'!\n", tag, m_name);
        return;
        }

    vi3x_help_funcs(&hf);

    if(NULL == rm->help)
        help_def(m_name, &hf);
    else
        rm->help(&hf);
    }

static void help_def(char const *m_name, vi3x_help_t const *hf)
    {
    hf->help_module(m_name, "-", "-", "-", "", NULL);
    hf->help_module_end();
    }

void vi3x_help_all_import(int help_short)
    {
    help_all(&list_import, help_short);
    }

void vi3x_help_all_export(int help_short)
    {
    help_all(&list_export, help_short);
    }

void vi3x_help_all_filter(int help_short)
    {
    help_all(&list_filter, help_short);
    }

static void help_all(module_list_t const *list, int help_short)
    {
    reg_module_t const *m;
    vi3x_help_t hf;

    for(m=list->first; NULL != m; m=m->next)
        {
        if(help_short)
            vi3x_help_short(&hf);
        else
            vi3x_help_funcs(&hf);
        if(NULL == m->help)
            help_def(m->name, &hf);
        else
            m->help(&hf);
        }
    }

static reg_module_t const *search_module(
    char const *m_name, module_list_t const *list, char const *tag)
    {
    char name[VI3X_NAME_SZ];
    char const *s;
    int len;
    reg_module_t const *rm;

    s=strpbrk(m_name, VI3X_D_END_RANGE);
    if(NULL == s)
        {
        len=strlen(m_name);
        s=m_name + len;
        }
    else
        len=s - m_name;
    if(0 == len)
        {
        vi3x_err("[%s] Error: Empty module name: '%s'!\n", tag, m_name);
        return NULL;
        }
    else if(len > VI3X_NAME_SZ - 1)
        {
        vi3x_err("[%s] Error: Module name too long: '%.*s'!\n", tag, len, m_name);
        return NULL;
        }
    memcpy(name, m_name, len);
    memset(name + len, 0, VI3X_NAME_SZ - len);

    for(rm=list->first; NULL != rm; rm=rm->next)
        {
        if(0 == strcmp(rm->name, name))
            return rm;
        }

    return NULL;
    }

int vi3x_is_import(vi3x_module_t const *m)
    {
    return match_module(&list_import, m);
    }

int vi3x_is_export(vi3x_module_t const *m)
    {
    return match_module(&list_export, m);
    }

static int match_module(module_list_t const *list, vi3x_module_t const *m)
    {
    reg_module_t const *rm;

    if(NULL == m)
        return 0;

    for(rm=list->first; NULL != rm; rm=rm->next)
        {
        if(rm->init == m->init)
            return 1;
        }

    return 0;
    }

static void vi3x_funcs(vi3x_t *vi3x)
    {
    *vi3x=Vi3x;
    }
