/*
This file is part of video3x.

video3x is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

video3x is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with video3x.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2011 Vadim Druzhin <cdslow@mail.ru>
*/

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "vi3x_types.h"
#include "vi3x_io.h"
#include "vi3x_utils.h"

static void insert_pair(long *range, long r0, long r1);
static void merge_ranges(long *range);

int vi3x_parse_args(
    void *params, char const *args, vi3x_arg_t const *map, char const *tag)
    {
    char *p=(char *)params;
    char const *name;
    char const *sep;
    char *colon;
    int name_len, val;
    vi3x_arg_t const *m;

    if(NULL == tag)
        tag="vi3x_parse_args";

    /* Assign deafult values */
    for(m=map; NULL != m->name; ++m)
        *(int *)(p + m->offset)=m->def;

    sep=args - 1;
    do
        {
        name=sep + 1;
        if(0 == *name)
            break;

        sep=strpbrk(name, VI3X_D_LIST_VAL);
        if(NULL == sep)
            sep=name + strlen(name);
        name_len=sep - name;

        if(VI3X_D_VAL == *sep)
            {
            ++sep;
            val=(int)strtol(sep, &colon, 10);
            if(colon != sep && (VI3X_D_LIST == *colon || 0 == *colon))
                sep=colon;
            else
                {
                sep=strchr(colon, VI3X_D_LIST);
                if(NULL == sep)
                    sep=colon + strlen(colon);
                vi3x_err("[%s] Error: Bad parameter value: '%.*s'\n",
                    tag, sep - name, name);
                return 0;
                }
            }
        else
            val=1;

        for(m=map; NULL != m->name; ++m)
            {
            if(0 == strncmp(m->name, name, name_len) && 0 == m->name[name_len])
                {
                *(int *)(p + m->offset)=val;
                break;
                }
            }

        if(NULL == m->name)
            {
            vi3x_err("[%s] Error: Unknown parameter: '%.*s'\n", tag, name_len, name);
            return 0;
            }
        }
    while(VI3X_D_LIST == *sep);

    return 1;
    }

void *vi3x_alloc(int sz, char const *tag)
    {
    void *ptr;

    ptr=calloc(sz, 1);
    if(NULL == ptr)
        vi3x_die_oom(tag);

    return ptr;
    }

void vi3x_free(void *ptr)
    {
    free(ptr);
    }

int vi3x_count_ranges(char const *args, char const *tag)
    {
    char const *str;
    char const *s;
    int n=0;

    if(NULL == tag)
        tag="vi3x_count_ranges";

    str=args;
    for(;;)
        {
        s=str;
        while(isdigit(*s))
            ++s;
        if(s == str)
            break;
        if(VI3X_D_PAIR == *s)
            {
            ++s;
            str=s;
            while(isdigit(*s))
                ++s;
            if(s==str)
                break;
            }
        ++n;
        if(*s != VI3X_D_LIST)
            {
            if(*s != 0 && *s != VI3X_D_END)
                break;
            return n * 2 + 1;
            }
        str=s + 1;
        }

    vi3x_err("[%s] Error: Invalid range specification: '%s'\n", tag, args);

    return 0;
    }

char const *vi3x_parse_ranges(long *range, char const *args, char const *tag)
    {
    char const *str;
    char *s;
    long r0, r1;

    if(NULL == tag)
        tag="vi3x_parse_ranges";

    *range=-1l;
    str=args;
    for(;;)
        {
        r0=strtol(str, &s, 10);
        if(s == str || r0 < 1l)
            break;
        if(VI3X_D_PAIR == *s)
            {
            str=s + 1;
            r1=strtol(str, &s, 10);
            if(s == str || r1 < r0)
                break;
            }
        else
            r1=r0;
        insert_pair(range, r0, r1);
        if(*s != VI3X_D_LIST)
            {
            if(*s != 0 && *s != VI3X_D_END)
                break;
            merge_ranges(range);
            return s;
            }
        str=s + 1;
        }

    vi3x_err("[%s] Error: Invalid range specification: '%s'\n", tag, args);

    return NULL;
    }

static void insert_pair(long *range, long r0, long r1)
    {
    long *rr;

    for(rr=range; *rr != -1l; rr+=2)
        {
        if(r1 + 1 < rr[0])
            /* Insert before */
            break;
        if(r0 - 1 <= rr[1])
            {
            /* Merge */
            if(rr[0] > r0)
                rr[0]=r0;
            if(rr[1] < r1)
                rr[1]=r1;
            return;
            }
        }

    for(range=rr; *rr != -1l; rr+=2)
        {}
    memmove(range + 2, range, (rr - range + 1) * sizeof(*range));
    range[0]=r0;
    range[1]=r1;
    }

static void merge_ranges(long *range)
    {
    long *rr;

    /* Check for empty list */
    if(-1l == *range)
        return;

    for(rr=range + 2; *rr != -1l; rr+=2)
        {
        if(rr[0] - 1 <= range[1])
            {
            /* Merge ranges */
            if(range[0] > rr[0])
                range[0]=rr[0];
            if(range[1] < rr[1])
                range[1]=rr[1];
            }
        else
            {
            range+=2;
            if(range < rr)
                {
                /* Move ranges */
                range[0]=rr[0];
                range[1]=rr[1];
                }
            }
        }
    range[2]=-1l;
    }

